/* im2avi_t.cxx copyright 2001, 2002 by Charl P. Botha http://cpbotha.net/
 *
 * $Id: im2avi_t.cxx,v 1.3 2001/12/28 17:20:52 cpbotha Exp $
 *
 * im2avi: frontend to avifile and Magick++ for making AVIs from images.
 * Copyright (C) 2001,2002  Charl P. Botha
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "event_codes.h"
#include "im2avi_t.h"
#include <iostream>
#include <fstream>
#include <math.h>
#include <strstream>

// includes for avifile
#include <avifile.h>
#include <aviplay.h>
#include <creators.h>
#include <except.h>
#include <fourcc.h>
#include <version.h>
#include <videoencoder.h>

// these have to come after avifile, else things break (class "frame" seems to be the culprit)
#include "FL/Fl.H"
#include "FL/Fl_File_Chooser.H"
#include "main_window_t.h"

// includes for Magick++
#include <Magick++.h>

// -----------------------------------------------------------------------
// global fltk callbacks
// -----------------------------------------------------------------------
void mw_button_cb(Fl_Button* w, void* user_data)
{
   // travel up the parent tree until we reach the top level window
   Fl_Widget *p = w;
   while (p->parent()) p = p->parent();
   // having found that, we can use the user_data to get hold of im2avi_t
   ((im2avi_t*)p->user_data())->mw_button_cb(w, user_data);
}

void mw_input_cb(Fl_Value_Input* w, void* user_data)
{
   // travel up the parent tree until we reach the top level window
   Fl_Widget *p = w;
   while (p->parent()) p = p->parent();
   // having found that, we can use the user_data to get hold of im2avi_t
   ((im2avi_t*)p->user_data())->mw_input_cb(w, user_data);
}

// -----------------------------------------------------------------------
// class im2avi_t definitions
// -----------------------------------------------------------------------
im2avi_t::im2avi_t(int argc, char* argv[])
{
   // instantiate, setup and show main window
   main_window = new main_window_t;
   main_window->window->user_data(this);
   main_window->window->show(argc, argv);

   // update the width and height internal variables according to UI
   update_image_width(true);
   update_image_height(true);
   // and set the initial aspect_ratio
   image_aspect_ratio = (float) image_width / (float) image_height;

   update_progress_ui(0);

   // instantiate the multiple selection file dialog for the loading of images
   file_chooser = new Fl_File_Chooser(".", "*", Fl_File_Chooser::MULTI, "Select images to add...");

   // this is a dummy create just so that the video_codecs list is filled in
   BITMAPINFOHEADER bih;
   bih.biCompression = 0xffffffff;
   Creators::CreateVideoDecoder(bih, 0, 0);

   // make sure our vector of fourccs is empty
   avi_codec_fourccs.clear();
   // use the simpler avifile vector from the avm namespace
   avm::vector<CodecInfo>::iterator it;
   // and go down the list of available compression codecs
   for (it=video_codecs.begin(); it!=video_codecs.end(); it++)
   {
      if (it->direction & CodecInfo::Encode)
      {
         // if it's an encoder, fill in the name in the codec browser
         main_window->avi_codec_choice->add(it->GetName());
         // and put the corresponding fourcc in the vector
         avi_codec_fourccs.push_back(it->fourcc);
      }
   }
   // default first codec (choice is 0-based)
   main_window->avi_codec_choice->value(0);
}

im2avi_t::~im2avi_t()
{
   // take care of the file_chooser
   delete file_chooser;
   // then nuke the main window
   delete main_window;
}

int im2avi_t::run(void)
{
   return Fl::run();
}

void im2avi_t::mw_button_cb(Fl_Button* w, void* user_data)
{
   switch (int(user_data))
   {
      case I2AE_ADD_IMAGES_FS: add_images_fs(); break;
      case I2AE_REMOVE_IMAGES: remove_images(); break;
      case I2AE_BROWSE_AVI_FILENAME: browse_avi_filename(); break;
      case I2AE_AUTO_WXH: auto_wxh(); break;
      case I2AE_MAKE_AVI: make_avi(); break;
      case I2AE_QUIT: exit(0); break;
      case I2AE_IMAGES_TOP: move_selection(1); break;
      case I2AE_IMAGES_UP: move_selection_diff(-1); break;
      case I2AE_IMAGES_DOWN: move_selection_diff(1); break;
      case I2AE_IMAGES_BOTTOM: move_selection_to_bottom(); break;
      case I2AE_LOAD_IMAGES_LIST: load_images_list(); break;
      case I2AE_SAVE_IMAGES_LIST: save_images_list(); break;
   }
}

void im2avi_t::mw_input_cb(Fl_Value_Input* w, void* user_data)
{
   switch (int(user_data))
   {
      case I2AE_WIDTH_CHANGED: update_image_width(); break;
      case I2AE_HEIGHT_CHANGED: update_image_height(); break;
   }
}

void im2avi_t::add_images_fs(void)
{
   file_chooser->show();

   while (file_chooser->shown())
   {
      Fl::wait();
   }

   // the parameter to file_chooser->value() is 1-based
   for (int i = 1; i <= file_chooser->count(); i++)
   {
      // add selected files to the browser
      main_window->images_browser->add(file_chooser->value(i));
   }
}

void im2avi_t::remove_images(void)
{
   // the Fl_Browser is 1-based; because we're removing items, we work
   // back to front, which makes everything much easier
   for (int i = main_window->images_browser->size(); i > 0; i--)
   {
      // if it's selected
      if (main_window->images_browser->selected(i))
      {
         // nuke it!
         main_window->images_browser->remove(i);
      }
   }
}

void im2avi_t::load_images_list(void)
{
   char *list_filename = fl_file_chooser("Select images list file to load", "*", NULL);
   if (list_filename)
   {
      ifstream list_file(list_filename);
      if (list_file)
      {
         char image_name[2048];
         // first get rid of current contents
         main_window->images_browser->clear();
         while (!list_file.eof())
         {
            list_file.getline(image_name, 2048);
            if (strlen(image_name) > 0)
            {
               main_window->images_browser->add(image_name);
            }
         } // while (!list_file.eof()) ...
      } // if (list_file) ...
      else
      {
         fl_message("Unable to open %s for reading!", list_filename);
      }
   } // if (list_filename) ...
}

void im2avi_t::save_images_list(void)
{
   char *list_filename = fl_file_chooser("Enter filename to save images list to", "*", NULL);
   if (list_filename)
   {
      ofstream list_file(list_filename);
      if (list_file)
      {
         for (int i = 1; i <= main_window->images_browser->size(); i++)
         {
            list_file << main_window->images_browser->text(i) << endl;
         }
      }
      else
      {
         fl_message("Unable to open %s for writing!", list_filename);
      }
   }
}

void im2avi_t::move_selection(int to)
{
   int first_selected = -1, number_selected = 0;
   for (int i = 1; i <= main_window->images_browser->size(); i++)
   {
      if (main_window->images_browser->selected(i))
      {
         if (first_selected < 0)
            first_selected = i;
         number_selected++;
      }
   }

   if (to < 1 || to > main_window->images_browser->size() - number_selected + 1)
   {
      cout << "im2avi_t::move_selection(): to == " << to << "!!" << endl;
      // something has gone wrong; this shouldn't happen.
      return;
   }
   
   if (first_selected > 0)
   {
      if (to < first_selected)
      {
         for (int i = 1; i <= main_window->images_browser->size(); i++)
         {
            if (main_window->images_browser->selected(i))
            {
               main_window->images_browser->move(i - (first_selected - to), i);
            }
         }
      } // if (to < first_selected) ...
      else if (to > first_selected)
      {
         for (int i = main_window->images_browser->size(); i > 0; i--)
         {
            if (main_window->images_browser->selected(i))
            {
               main_window->images_browser->move(i + (to - first_selected), i);
            }
         }
      } // else if (to > first_selected) ...
   } // if (first_selected > 0) ...
}

void im2avi_t::move_selection_diff(int diff)
{
   int first_selected = -1;
   // CAREFUL: we abort the moment first_selected is set; if this changes, make sure
   // first_selected isn't corrupted
   for (int i = 1; i <= main_window->images_browser->size() && first_selected < 0; i++)
   {
      if (main_window->images_browser->selected(i))
         first_selected = i;
   }
   if (first_selected > 0)
   {
      move_selection(first_selected + diff);
   }
}

void im2avi_t::move_selection_to_bottom(void)
{
   int number_selected = 0;
   for (int i = 1; i <= main_window->images_browser->size(); i++)
   {
      if (main_window->images_browser->selected(i))
         number_selected++;
   }

   move_selection(main_window->images_browser->size() - number_selected + 1);
}

void im2avi_t::update_image_width(bool ignore_aspect)
{
   image_width = (int)main_window->im_width_input->value();
   main_window->wm4_output->value(image_width % 4);
   main_window->wm8_output->value(image_width % 8);
   main_window->wm16_output->value(image_width % 16);
   if (!ignore_aspect && main_window->constrain_aspect_check->value())
   {
      main_window->im_height_input->value(rint(image_width / image_aspect_ratio));
      update_image_height(true);
   }
}

void im2avi_t::update_image_height(bool ignore_aspect)
{
   image_height = (int)main_window->im_height_input->value();
   if (!ignore_aspect && main_window->constrain_aspect_check->value())
   {
      main_window->im_width_input->value(rint(image_height * image_aspect_ratio));
      update_image_width(true);
   }
}

void im2avi_t::browse_avi_filename(void)
{
   char *avi_filename = fl_file_chooser("Enter AVI filename", "*.avi", NULL);
   if (avi_filename)
   {
      main_window->avi_filename_input->value(avi_filename);
   }
}

void im2avi_t::auto_wxh(void)
{
   bool found = false;
   int i;
   // search for the first selected image
   for (i = 0; i < main_window->images_browser->size() && !found; i++)
   {
      found = main_window->images_browser->selected(i+1);
   }
   // get size from selected image
   if (found)
   {
      // i was zero-based, but was incremented at the end of the for loop, so now is 1-based
      try
      {
         // open the image
         Magick::Image image(main_window->images_browser->text(i));
         // and attempt to enter its dimensions into the applicable input boxes
         // also update internal variables, but don't pay attention to "constrain aspect ratio" setting
         main_window->im_width_input->value(image.columns());
         update_image_width(true);
         main_window->im_height_input->value(image.rows());
         update_image_height(true);
         image_aspect_ratio = (float) image_width / (float) image_height;
      }
      catch (Magick::Exception &error_)
      {
         // sumfink went wronk
         fl_message("Failed to open first image: %s", error_.what());
      }
   }
   else
   {
      fl_message("You have to select an image for me to use!");
   }
}

void im2avi_t::make_avi(void)
{
   if (strlen(main_window->avi_filename_input->value()) <= 0 || main_window->images_browser->size() <= 0)
   {
      fl_message("You have to have images in the images list and an output AVI filename.");
      return;
   }

   bool exception_caught = false;
   IAviWriteFile* avi_file = NULL;
   IAviVideoWriteStream* stream = NULL;
   int width = image_width;
   int height = image_height;
   BitmapInfo bh(width, height, 24);

   update_progress_ui(0);

   try
   {
      avi_file = CreateIAviWriteFile(main_window->avi_filename_input->value());
      // this tells avifile about our input images (width, height and RGB for now)
      stream = avi_file->AddVideoStream(avi_codec_fourccs[main_window->avi_codec_choice->value()], 
                                                              &bh, int(1000000.0 / main_window->fps_input->value()) );
      // setup quality
      stream->SetQuality(100 * main_window->quality_slider->value());
      // and away we go!
      stream->Start();
   }
   catch (FatalError& e)
   {
      exception_caught = true;
      if (avi_file)
      {
         stream->Stop();
         delete avi_file;
      }
      fl_message("Unable to create AVI stream: %s", e.GetDesc());
      // stop this function; we've deallocated everything we could
      return;
   }

   // allocate some temporary bitmap buffer
   unsigned char *image_data = new unsigned char[width * height * 3];
   try
   {
      for (int i = 0; i < main_window->images_browser->size(); i++)
      {
         // FL_Browsers are 1-based *sigh*
         // open the image
         Magick::Image image(main_window->images_browser->text(i + 1));
         // we have to flip it vertically (because avifile seems to like it that way)
         image.flip();
         // and we have to resize it to the desired width and height
         Magick::Geometry new_geom(width, height);
         // but we ignore aspect, as we take care of this ourselves
         new_geom.aspect(true);
         image.zoom(new_geom);
         // write out whole image to buffer as RGB chars
         image.write(0, 0, width, height, "BGR", Magick::CharPixel, image_data);
         // create a CImage from this, but don't copy the data
         CImage avi_image(&bh, image_data, false);
         // and add it to the stream
         stream->AddFrame(&avi_image);
         // update the progress thingy
         update_progress_ui((i + 1) / (float) main_window->images_browser->size() * 100.0);
         // and flush all ui updates! (else nothing gets redrawn while we process)
         Fl::flush();
      }
   }
   catch (FatalError& e)
   {
      fl_message("Unable to write to AVI stream: %s", e.GetDesc());
   }
   catch (Magick::Exception &error_)
   {
      // sumfink went wronk
      fl_message("Unable to process input image during encoding: %s", error_.what());
   }

   // nuke the bitmap buffer
   delete[] image_data;
   // make sure the stream is stopped
   stream->Stop();
   // finish the file
   if (avi_file)
   {
      delete avi_file;
   }
}

void im2avi_t::update_progress_ui(float percentage)
{
   static string percentage_string;
   
   // set progress bar to the new value
   main_window->avi_build_progress->value(percentage);
   
   // use stringstream to make a suitable label for the progress bar
   strstream pss;
   pss << percentage << " %" << ends;
   // store the result in our static string ("label" just copies the pointer!)
   percentage_string = pss.str();
   // make sure all the mem associated with the strstream is deallocated at its destruction
   pss.freeze(0);

   // finally set the label
   main_window->avi_build_progress->label(percentage_string.c_str());

   // make sure this knows it has to be redrawn (label()s are often inlines,
   // they are too fast to be doing redraw()s)
   main_window->avi_build_progress->redraw();
}

// this doesn't have to global, but it's convenient if we add modules later
im2avi_t* im2avi;

int main(int argc, char* argv[])
{
   im2avi = new im2avi_t(argc, argv);
   int run_result = im2avi->run();
   delete im2avi;
   return run_result;
}
